# -*- coding: utf-8 -*-

import os
from shutil import copyfile, copytree
import pkg_resources
from .__init__ import __version__, __source_repository__, __release_repository__
import configparser
import logging
from pykwalify.core import Core
import pykwalify
import glob
import yaml
import sys


class ConfigValidator:
    def __init__(self):
        self.logging = logging
        self.paths = dict()
        self.plugins = str()
        self.logging_level = str()
        self.reader = None
        self.configuration = dict()

        # Figure out where we're installed and set defaults
        self.paths['bin_dir'] = os.path.dirname(os.path.abspath(__file__))
        self.paths['root_dir'] = os.path.dirname(self.paths['bin_dir'])

    @staticmethod
    def get_version_information():
        version_information = """
            Application Converge
            Version: {version_number:s}
            Project Source: {source_repository:s}
            Release repository: {release_repository:s}
            """
        version_arguments = {
            "version_number": __version__,
            "source_repository": __source_repository__,
            "release_repository": __release_repository__
        }
        return version_information.format(**version_arguments)

    @staticmethod
    def init_conf(target_directory):
        result = False
        init_path = os.path.isdir(os.path.join(target_directory, "converge.ini.template"))
        if not init_path:
            print("will create configuration file in %s/converge.ini.template" % target_directory)
            resource_package = __name__
            resource_path = os.path.join('resources', 'etc')
            template = pkg_resources.resource_filename(resource_package, resource_path)
            print("Copying template from %s to %s" % (template, target_directory))
            copytree(template, target_directory)
            print("New configuration can be found in: %s" % target_directory)
            print("Make you modifications and rename it to converge.ini")
            result = os.path.isfile(os.path.join(target_directory, "converge.ini.template"))
        else:
            print("File already exists: %s/converge.ini.template" % target_directory)
        return result

    @staticmethod
    def get_directory_tree(directory_path):
        tree = []
        for dirname, dirnames, filenames in os.walk(directory_path):
            for subdirname in dirnames:
                tree.append(os.path.join(dirname, subdirname))
            for filename in filenames:
                tree.append(os.path.join(dirname, filename))
        return tree

    def init_repository(self, target_directory):
        result = False
        init_path = os.path.isdir(target_directory)
        if not init_path:
            print("will copy repository template into %s" % target_directory)
            resource_package = __name__
            resource_path = os.path.join('resources', 'repository')
            template = pkg_resources.resource_filename(resource_package, resource_path)
            print("Copying contents of repository template from %s to %s" % (template, target_directory))
            copytree(template, target_directory)
            print("Newly initialized repository can be found in: %s" % target_directory)
            dir_tree = self.get_directory_tree(target_directory)
            if isinstance(dir_tree, list) and len(dir_tree) > 0:
                result = True
        else:
            print("Folder already exists: %s" % target_directory)
        return result

    def load_config(self,config_path):
        result = False
        path_exists = os.path.isfile(config_path)
        if path_exists:
            print("Checking configuration at location: '%s'" % config_path)
            converge_schema_path = os.path.join(self.paths["bin_dir"], "schemas", "converge_schema.yaml")
            if self.validate_yaml_schema(target_path=os.path.dirname(config_path), schema_path=converge_schema_path):
                with open(config_path, 'r') as stream:
                    try:
                        self.configuration = yaml.load(stream)
                        result = True
                    except yaml.YAMLError as exc:
                        print(exc)
        else:
            print("File %s does not exist" % config_path)
        return result

    def check_config(self, config_path):
        result = self.load_config(config_path=config_path)

        if result:
            print("\t## Configuration to be used:\n")
            print("\tLogging Level: '%s'" % self.configuration['default']['logging_level'])
            print("\tAvailable programs: %s" % ", ".join(self.configuration['programs'].keys()))
            for prog in self.configuration['programs']:
                print("\n\t# \"%s\" Program Comfiguration:" % prog)
                for conf, subconfs in self.configuration['programs'][prog]['conf'].items():
                    for name, sub in subconfs.items():
                        print("\t%s : %s : %s" % (conf, name, sub))
            print("")

        return result

    def validate_hierarchy_yaml(self):
        target_path = self.paths['hierarchy']
        schema_path = os.path.join(self.paths['bin_dir'], "schemas", "hierarchy_schema.yaml")
        result = self.validate_yaml_schema(target_path=target_path, schema_path=schema_path)
        return result

    def validate_host_yaml(self):
        target_path = self.paths['host']
        schema_path = os.path.join(self.paths['bin_dir'], "schemas", "host_schema.yaml")
        result = self.validate_yaml_schema(target_path=target_path, schema_path=schema_path)
        return result

    def validate_application_yaml(self):
        target_path = self.paths['application']
        schema_path = os.path.join(self.paths['bin_dir'], "schemas", "application_schema.yaml")
        result = self.validate_yaml_schema(target_path=target_path, schema_path=schema_path)
        return result

    def validate_yaml_schema(self, target_path, schema_path):
        result = True
        self.logging.info("SCANNING: %s/**.yaml" % target_path)
        for filename_path in glob.iglob(os.path.join(target_path, "*.yaml")):
            self.logging.debug("Validating: %s" % filename_path)
            c = Core(source_file=filename_path, schema_files=[schema_path])
            try:
                c.validate(raise_exception=True)
            except pykwalify.errors.SchemaError:
                result = False
        if result:
            self.logging.info("VALIDATED: %s/**.yaml" % target_path)
        else:
            self.logging.error("Corrupt files!")
        return result

    def check_repository(self):
        result = True

        returns = self.validate_hierarchy_yaml()
        if returns is not True:
            result = False

        returns = self.validate_host_yaml()
        if returns is not True:
            result = False

        returns = self.validate_application_yaml()
        if returns is not True:
            result = False

        return result

    def get_configuration_paths(self):
        return self.paths



# # -*- coding: utf-8 -*-
#
# import pytest
# import unittest
# import sys
# import os
# sys.path.append(os.path.abspath(os.path.join('..', 'converge')))
# from pyconverge.Helpers import Helpers as Helpers
#
#
# # class TestHelpers(unittest.TestCase):
# #
# #     def setUp(self):
# #         self.helpers = Helpers()
# #
# #     def test_get_directory_tree_exists(self):
# #         result = False
# #         directory_path = "pyconverge/resources"
# #         returned = self.helpers.get_directory_tree(directory_path=directory_path)
# #         print(returned)
# #         if isinstance(returned, list) and returned is not None and len(returned) > 1:
# #             result = True
# #         self.assertTrue(result)
# #
# #     def test_get_directory_tree_not_exists(self):
# #         result = False
# #         directory_path = "pyconverge/resources_not_exists"
# #         returned = self.helpers.get_directory_tree(directory_path=directory_path)
# #         print(returned)
# #         if isinstance(returned, list) and returned == []:
# #             result = False
# #         self.assertFalse(result)
# #
# # if __name__ == '__main__':
# #     unittest.main()


# # -*- coding: utf-8 -*-
#
# import pytest
# import unittest
# import sys
# import os
#
# sys.path.append(os.path.abspath(os.path.join('..', 'converge')))
# from pyconverge.RepositoryReader import RepositoryReader
#
#
# class TestRepositoryReader(unittest.TestCase):
#     def setUp(self):
#         self.repository_path = "tests/resources/repository"
#         self.node_path = "tests/resources/repository/nodes"
#         self.node_group_path = "tests/resources/repository/node_groups"
#         self.package_path = "tests/resources/repository/packages"
#         self.package_group_path = "tests/resources/repository/package_groups"
#         self.hierarchy_path = "tests/resources/repository/hierarchy"
#         self.package_inheritance_depth_max = 7
#
#         args = {
#             "hierarchy_path": self.hierarchy_path,
#             "repository_path": self.repository_path,
#             "node_path": self.node_path,
#             "node_group_path": self.node_group_path,
#             "package_path": self.package_path,
#             "package_group_path": self.package_group_path,
#             "package_inheritance_depth_max": self.package_inheritance_depth_max
#         }
#         self.repositoryreader = RepositoryReader(**args)
#
#     def test_load_yaml_files_in_directory(self):
#         result = False
#         returns = self.repositoryreader.load_yaml_files_in_directory(self.hierarchy_path)
#         if isinstance(returns, dict) and 'nodes::nodes' in returns['hierarchy']['node::hierarchy']:
#             result = True
#         self.assertTrue(result)
#
#     def test_load_non_resolved_configuration(self):
#         result = False
#         returns = self.repositoryreader.load_non_resolved_configuration()
#         if isinstance(returns, dict) \
#                 and returns["hierarchy"] \
#                 and returns["nodes"] \
#                 and returns["node_groups"] \
#                 and returns["packages"] \
#                 and returns["package_groups"]:
#             result = True
#         self.assertTrue(result)
#
#     def test_validate_yaml_schema(self):
#         result = False
#         args = {
#             "target_path": self.hierarchy_path,
#             "schema_path": os.path.join("pyconverge", "schemas", "hierarchy_schema.yaml")
#         }
#         returns = self.repositoryreader.validate_yaml_schema(**args)
#         if returns is True:
#             result = True
#         self.assertTrue(result)
#
#     def test_validate_node_yaml(self):
#         result = False
#         returns = self.repositoryreader.validate_node_yaml()
#         if returns is True:
#             result = True
#         self.assertTrue(result)
#
#     def test_validate_node_group_yaml(self):
#         result = False
#         returns = self.repositoryreader.validate_node_group_yaml()
#         if returns is True:
#             result = True
#         self.assertTrue(result)
#
#     def test_validate_package_yaml(self):
#         result = False
#         returns = self.repositoryreader.validate_package_yaml()
#         if returns is True:
#             result = True
#         self.assertTrue(result)
#
#     def test_validate_package_group_yaml(self):
#         result = False
#         returns = self.repositoryreader.validate_package_group_yaml()
#         if returns is True:
#             result = True
#         self.assertTrue(result)
#
#     def test_validate_hierarchy_yaml(self):
#         result = False
#         returns = self.repositoryreader.validate_hierarchy_yaml()
#         if returns is True:
#             result = True
#         self.assertTrue(result)


# -*- coding: utf-8 -*-

from importlib import import_module
import logging

log = logging.getLogger(__name__)


# Dynamically import libraries based on user preferences
def get_dynamic_class(finder_path):
    module_name, class_name = finder_path.rsplit('.', 1)
    module = import_module(module_name)
    log.debug("Imported Module %s, Class %s" % (module_name, class_name))
    return getattr(module, class_name)


class ConvergeData(object):
    data = dict()
    targets = dict()


class BaseClassLoader:
    def __init__(self, settings):
        self.programs = settings["programs"]
        self.settings = dict()
        self.instructions = list()

    def run_instruction_set(self, **kwargs):
        result = False
        data = ConvergeData()
        program_name = kwargs.get("program")
        mode = kwargs.get("mode")
        arguments = kwargs.get("arguments")
        settings = kwargs.get("settings")

        self.instructions = self.programs[program_name]["modes"][mode]
        self.settings = self.programs[program_name]["conf"]

        for instruction in self.instructions:
            dynamic_class_path = instruction
            runner_class = get_dynamic_class(finder_path=dynamic_class_path)
            runner = runner_class()

            data = runner.run(data=data, conf=settings, **arguments)
        return result



# -*- coding: utf-8 -*-

import logging


class PrintHostTags:
    @staticmethod
    def run(data, **kwargs):
        for host_name in data.targets["hosts"]:
            message = "HOST TAG LOOKUP \n %s tags:\n\t%s"
            if host_name in data.data["hosts"]:
                host_tags = data.data["hosts"][host_name]
                logging.info(message % (host_name, str(host_tags)))
        return data


class PrintHostApplications:
    @staticmethod
    def run(data, **kwargs):
        message = "HOST TO APPLICATION LOOKUP \n HOST: %s has applications:\n\t%s"
        host_name = kwargs.get("host_name")
        logging.info(message % (host_name, data.targets["applications"]))
        return data


class PrintTagHosts:
    @staticmethod
    def run(data, **kwargs):
        message = "TAG TO HOST LOOKUP \n TAG: %s=%s has hosts:\n\t%s"
        tag_name = kwargs.get("tag_name")
        tag_value = kwargs.get("tag_value")
        logging.info(message % (tag_name, tag_value, list(data.targets["hosts"])))
        return data


class PrintTagApplications:
    @staticmethod
    def run(data, **kwargs):
        message = "TAG TO APPLICATION LOOKUP \n TAG: %s=%s has applications:\n\t%s"
        tag_name = kwargs.get("tag_name")
        tag_value = kwargs.get("tag_value")
        logging.info(message % (tag_name, tag_value, data.targets["applications"]))
        return data


class PrintApplicationHosts:
    @staticmethod
    def run(data, **kwargs):
        message = "APPLICATION TO HOST LOOKUP \n APPLICATION: %s has hosts:\n\t%s"
        application_name = kwargs.get("application_name")
        logging.info(message % (application_name, data.targets["hosts"]))
        return data


class PrintApplicationTags:
    @staticmethod
    def run(data, **kwargs):
        message = "APPLICATION TO TAG LOOKUP \n APPLICATION: %s has tags:\n\t%s"
        application_name = kwargs.get("application_name")
        logging.info(message % (application_name, data.data["application_hosts"][application_name]))
        return data


class PrintApplicationProperties:
    @staticmethod
    def run(data, **kwargs):
        message = "APPLICATION TO PROPERTIES LOOKUP \n APPLICATION: %s has properties:\n\t%s"
        appliation_name = kwargs.get("application_name")
        logging.info(message % (appliation_name, list(data.data["application_properties"][appliation_name]["properties"])))
        return data


class PrintPropertyApplications:
    @staticmethod
    def run(data, **kwargs):
        message = "PROPERTIES TO APPLICATION LOOKUP \n PROPERTIES: %s has applications:\n\t%s"
        property_name = kwargs.get("property_name")
        logging.info(message % (property_name, data.targets["applications"]))
        return data



# -*- coding: utf-8 -*-

import os


class OutputProperties:

    def __init__(self, output_path="output"):
        self.output_path = output_path

    def generate_files_by_target(self, configuration):
        result = False
        for host_name, applications in configuration.items():
            for application_name, properties in applications.items():
                output_folder = os.path.join(self.output_path, host_name, application_name)
                if not os.path.exists(output_folder):
                    os.makedirs(output_folder)
                for property_name, keys in properties.items():
                    file_name = "%s.properties" % property_name
                    for key in keys:
                        with open(os.path.join(output_folder, file_name), 'w+') as out:
                            out.write("%s = %s" % (key[0], key[1]))

        return result



# -*- coding: utf-8 -*-

import argparse
import sys
import logging


class ArgumentParser:
    def __init__(self, **kwargs):
        self.configuration = kwargs.get("configuration")

    @staticmethod
    def add_sub_parser_group(sp, option_name, config):
        description = str()
        parser_group = argparse.ArgumentParser(add_help=False)
        for option in config["conf"]["default"]["args"]:
            parser_group.add_argument(option, action="store", type=str, default=None)
        options = config["modes"].keys()
        if "description" in config["conf"]["default"]:
            description = config["conf"]["default"]["description"]
        parser_group.add_argument("mode", action="store", choices=options)
        sp_new = sp.add_parser(option_name, parents=[parser_group], help=description)
        sp_new.set_defaults(which=option_name)
        return sp

    def create_parser(self):
        parser = argparse.ArgumentParser()

        # # option groups to do:
        # tag
        # host
        # application
        # property
        # hierarchy

        parser.add_argument("-v", "--verbose",
                            action="store_true", default=False, required=False,
                            help="run program in verbose/debug mode, lots of output!")
        parser.add_argument("--stdout",
                            action="store", choices=["INFO", "WARNING", "ERROR", "CRITICAL", "DEBUG"],
                            default=["WARNING"], required=False,
                            help="change stdout logging level (logs INFO to file already)")

        parser.add_argument("--config", action="store",
                            type=str, default="converge.yaml",
                            help="path to the configuration file")

        group_init = argparse.ArgumentParser(add_help=False)
        group_init.add_argument("init_type", action="store",
                                choices=["repository", "conf"],
                                help="choose to initialize repository or conf")
        group_init.add_argument("path", action="store",
                                type=str, default=None,
                                help="this path will be the initialization root")

        group_checkconfig = argparse.ArgumentParser(add_help=False)

        group_version = argparse.ArgumentParser(add_help=False)
        group_version.add_argument("--version", action="store_true", default=True, required=False)

        # activate subparsers on main parser
        sp = parser.add_subparsers()

        sp_init = sp.add_parser("init", parents=[group_init], help="initialize configuration or repository")
        sp_init.set_defaults(which="init")

        sp_checkconfig = sp.add_parser("check", parents=[group_checkconfig],
                                       help="run sanity check on configuration")
        sp_checkconfig.set_defaults(which="check")

        sp_version = sp.add_parser("version", parents=[group_version],
                                   help="get converge version and build information")
        sp_version.set_defaults(which="version")

        """ AUTOMATIC OPTION LOADING """
        config_path = "converge.yaml"
        for i, arg in enumerate(sys.argv):
            if arg == "--config" and len(sys.argv) > i + 1:
                config_path = sys.argv[i + 1]
                break

        result = self.configuration.load_config(config_path=config_path)
        if result:
            logging.info("OK: Configuration file %s" % config_path)
            programs = self.configuration.configuration["programs"]
            for program_name, program_config in programs.items():
                sp = self.add_sub_parser_group(sp=sp, option_name=program_name, config=program_config)

        return parser



# -*- coding: utf-8 -*-

from pyconverge.ArgumentParser import ArgumentParser
from pyconverge.ConfigValidator import ConfigValidator
from .BaseClassLoader import BaseClassLoader
import time
import logging
import sys
logging.getLogger("pykwalify.core").setLevel(logging.WARN)
logging.basicConfig(level="WARN")


# add main entry point
def main():
    statistics = dict()
    statistics['start_time'] = time.time()

    configuration = ConfigValidator()

    parser = ArgumentParser(configuration=configuration).create_parser()
    args = parser.parse_args()

    # initialize logging level
    try:
        logging.getLogger().setLevel(configuration.configuration["default"]["logging_level"])
    except:
        pass

    try:
        # version option
        if hasattr(args, "which"):
            if args.which == "version" and hasattr(args, "version"):
                statistics['opt_version'] = time.time()
                result = configuration.get_version_information()
                print(result)
                statistics['opt_version'] = time.time() - statistics['opt_version']

            # init options
            elif args.which == "init" and hasattr(args, "init_type") and args.init_type == "conf":
                statistics['opt_init_conf'] = time.time()
                configuration.init_conf(target_directory=args.path)
                statistics['opt_init_conf'] = time.time() - statistics['opt_init_conf']
            elif args.which == "init" and hasattr(args, "init_type") and args.init_type == 'repository':
                statistics['opt_init_repo'] = time.time()
                configuration.init_repository(target_directory=args.path)
                statistics['opt_init_repo'] = time.time() - statistics['opt_init_repo']

            # sanity check / check config
            elif args.which == "check" and hasattr(args, "config"):
                statistics['opt_check'] = time.time()
                result = configuration.check_config(config_path=args.config)
                if result:
                    # todo: add validator parsing from yaml file loaded above (dynamic checks)
                    logging.info("OK: Configuration file %s" % args.config)
                statistics['opt_check'] = time.time() - statistics['opt_check']
            elif args.which and hasattr(args, "config"):
                statistics[args.which] = time.time()
                settings = configuration.configuration
                mode = args.mode
                class_loader = BaseClassLoader(settings=settings)
                class_loader.run_instruction_set(program=args.which, mode=mode, settings=settings, arguments=vars(args))
                statistics[args.which] = time.time() - statistics[args.which]
            else:
                print("Application exited prematurely, options passed without error but nothing happened!")
                sys.exit(1)
        # statistics calculations
        statistics["end_time"] = time.time()
        statistics["total_time"] = statistics["end_time"] - statistics['start_time']
        logging.debug(statistics)
        logging.debug("Time elapsed: %f" % statistics["total_time"])
        return True
    except:
        # statistics calculations
        statistics["end_time"] = time.time()
        statistics["total_time"] = statistics["end_time"] - statistics['start_time']

        logging.debug("Time elapsed: %f" % statistics["total_time"])
        raise



# -*- coding: utf-8 -*-

from .LoadDataFromDisk import LoadDataFromDisk
import yaml
import os
import logging


class LoadHosts(LoadDataFromDisk):
    def merge_contents_of_files(self, file_list):
        contents = dict()
        all_yaml_individual = map(lambda x: yaml.load_all(open(x)), file_list)
        for configs in all_yaml_individual:
            for config in configs:
                contents = {**contents, **config}
        return contents

    def load_contents_of_files(self, glob_pattern=None):
        file_list = self.get_list_of_files(glob_pattern=glob_pattern, recursive=False)
        content = self.merge_contents_of_files(file_list=file_list)
        return content

    def run(self, data, conf, **kwargs):
        base_dir = conf["programs"]["host"]["conf"]["properties"]["base_dir"]
        host_glob = conf["programs"]["host"]["conf"]["properties"]["host_glob"]
        glob_pattern = os.path.join(base_dir, host_glob)
        data.data["hosts"] = self.load_contents_of_files(glob_pattern=glob_pattern)
        data.targets["hosts"] = list(data.data["hosts"].keys())
        return data


class LoadApplicationHostMapping(LoadDataFromDisk):
    def merge_contents_of_files(self, file_list):
        contents = dict()
        for file_path in file_list:
            app_name = file_path.rsplit("/", 1)[1][:-5]
            for config in yaml.load_all(open(file_path)):
                contents[app_name] = config
        return contents

    def load_contents_of_files(self, glob_pattern=None):
        file_list = self.get_list_of_files(glob_pattern=glob_pattern, recursive=False)
        content = self.merge_contents_of_files(file_list=file_list)
        return content

    def run(self, data, conf, **kwargs):
        base_dir = conf["programs"]["host"]["conf"]["properties"]["base_dir"]
        host_mapping_glob = conf["programs"]["host"]["conf"]["properties"]["host_mapping_glob"]
        glob_pattern = os.path.join(base_dir, host_mapping_glob)
        data.data["application_hosts"] = self.load_contents_of_files(glob_pattern=glob_pattern)
        data.targets["applications"] = set(data.data["application_hosts"].keys())
        return data


class LoadApplicationPropertiesMapping(LoadDataFromDisk):
    def merge_contents_of_files(self, file_list):
        contents = dict()
        for file_path in file_list:
            app_name = file_path.rsplit("/", 3)[1]
            for config in yaml.load_all(open(file_path)):
                contents[app_name] = config
        return contents

    def load_contents_of_files(self, glob_pattern=None):
        file_list = self.get_list_of_files(glob_pattern=glob_pattern, recursive=False)
        content = self.merge_contents_of_files(file_list=file_list)
        return content

    def run(self, data, conf, **kwargs):
        base_dir = conf["programs"]["application"]["conf"]["properties"]["base_dir"]
        property_mapping_glob = conf["programs"]["application"]["conf"]["properties"]["property_mapping_glob"]
        glob_pattern = os.path.join(base_dir, property_mapping_glob)
        data.data["application_properties"] = self.load_contents_of_files(glob_pattern=glob_pattern)
        data.targets["applications"] = set(data.data["application_properties"].keys())
        return data


class LoadApplications(LoadDataFromDisk):
    def merge_contents_of_files(self, file_list):
        contents = dict()
        for file_path in file_list:
            app_name = file_path.rsplit("/", 1)[1][:-5]
            for config in yaml.load_all(open(file_path)):
                contents[app_name] = config
        return contents

    def load_contents_of_files(self, base_directory="."):
        glob_pattern = os.path.join(base_directory, "applications", "*.yaml")
        file_list = self.get_list_of_files(glob_pattern=glob_pattern, recursive=False)
        content = self.merge_contents_of_files(file_list=file_list)
        return content



# -*- coding: utf-8 -*-

import pytest
import unittest
import sys
import os
import argparse
sys.path.append(os.path.abspath(os.path.join('..', 'converge')))
from pyconverge.ArgumentParser import ArgumentParser as ArgumentParser


class TestArgumentParser(unittest.TestCase):

    def setUp(self):
        self.argumentparser = ArgumentParser()

    def test_create_parser(self):
        result = False
        returns = self.argumentparser.create_parser()
        print(type(returns))
        if isinstance(returns, argparse.ArgumentParser):
            result = True
        self.assertTrue(result)




# -*- coding: utf-8 -*-

from abc import ABC, abstractmethod
import glob


class LoadDataFromDisk(ABC):
    @staticmethod
    def get_list_of_files(glob_pattern, recursive=False):
        return glob.glob(glob_pattern, recursive=recursive)

    @abstractmethod
    def merge_contents_of_files(self, file_list):
        pass

    @abstractmethod
    def read_data(self):
        pass



# -*- coding: utf-8 -*-

import pytest
import argparse
import unittest
import sys
import os

sys.path.append(os.path.abspath(os.path.join('..', 'converge')))
from pyconverge import converge as converge
import pyconverge.__main__ as mainexec


class TestConverge(unittest.TestCase):

    def test_main_arg_parser(self):
        try:
            result = converge.main()
        except argparse.ArgumentError:
            result = True
        self.assertTrue(result)

    def test_package_executable_main(self):
        try:
            result = mainexec.main()
        except argparse.ArgumentError:
            result = True
        self.assertTrue(result)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

import pytest
import unittest
import sys
import os

sys.path.append(os.path.abspath(os.path.join('..', 'converge')))
from pyconverge.BaseClassLoader import ConvergeData
import pyconverge.plugins.properties.FilterTargets as FilterTargets


class TestFilterTargets(unittest.TestCase):
    def setUp(self):
        self.data = ConvergeData()
        self.data.data = {
            'hosts': {
                'pre-host2': {'datacenter': 'ONE', 'network': 'PRE', 'pool': ['hostgroup2'], 'environment': 'pre'},
                'staging-host1': {'datacenter': 'ONE', 'network': 'DMZ', 'pool': ['hostgroup1'],
                                  'environment': 'staging'},
                'prod-host10': {'datacenter': 'TWO', 'network': 'SEC', 'pool': ['hostgroup4'], 'environment': 'prod'},
                'prod-host2': {'datacenter': 'TWO', 'network': 'DMZ', 'pool': ['hostgroup1'], 'environment': 'prod'},
                'prod-host7': {'datacenter': 'ONE', 'network': 'DMZ', 'pool': ['hostgroup2'], 'environment': 'prod'},
                'pre-host1': {'datacenter': 'ONE', 'network': 'PRE', 'pool': ['hostgroup1'], 'environment': 'pre'},
                'uat-host2': {'datacenter': 'ONE', 'network': 'UAT', 'pool': ['hostgroup2'], 'environment': 'uat'},
                'dev-host3': {'datacenter': 'ONE', 'network': 'DEV', 'pool': ['hostgroup3'], 'environment': 'dev'},
                'prod-host1': {'datacenter': 'ONE', 'network': 'DMZ', 'pool': ['hostgroup1'], 'environment': 'prod'},
                'local-host1': {'datacenter': None, 'network': 'local', 'pool': ['hostgroup3'], 'environment': 'local'},
                'prod-host3': {'datacenter': 'ONE', 'network': 'DMZ', 'pool': ['hostgroup1'], 'environment': 'prod'},
                'prod-host11': {'datacenter': 'ONE', 'network': 'SEC', 'pool': ['hostgroup4'], 'environment': 'prod'},
                'uat-host4': {'datacenter': 'ONE', 'network': 'UAT', 'pool': ['hostgroup4'], 'environment': 'uat'},
                'staging-host3': {'datacenter': 'ONE', 'network': 'SEC', 'pool': ['hostgroup3'],
                                  'environment': 'staging'},
                'pre-host4': {'datacenter': 'ONE', 'network': 'PRE', 'pool': ['hostgroup4'], 'environment': 'pre'},
                'prod-host6': {'datacenter': 'TWO', 'network': 'DMZ', 'pool': ['hostgroup2'], 'environment': 'prod'},
                'dev-host1': {'datacenter': 'ONE', 'network': 'DEV', 'pool': ['hostgroup1'], 'environment': 'dev'},
                'staging-host4': {'datacenter': 'ONE', 'network': 'SEC', 'pool': ['hostgroup4'],
                                  'environment': 'staging'},
                'staging-host2': {'datacenter': 'ONE', 'network': 'DMZ', 'pool': ['hostgroup2'],
                                  'environment': 'staging'},
                'pre-host3': {'datacenter': 'ONE', 'network': 'PRE', 'pool': ['hostgroup3'], 'environment': 'pre'},
                'dev-host4': {'datacenter': 'ONE', 'network': 'DEV', 'pool': ['hostgroup4'], 'environment': 'dev'},
                'uat-host1': {'datacenter': 'ONE', 'network': 'UAT', 'pool': ['hostgroup1'], 'environment': 'uat'},
                'uat-host3': {'datacenter': 'ONE', 'network': 'UAT', 'pool': ['hostgroup3'], 'environment': 'uat'},
                'prod-host12': {'datacenter': 'TWO', 'network': 'SEC', 'pool': ['hostgroup4'], 'environment': 'prod'},
                'local-host2': {'datacenter': None, 'network': 'local', 'pool': ['hostgroup4'], 'environment': 'local'},
                'prod-host8': {'datacenter': 'TWO', 'network': 'DMZ', 'pool': ['hostgroup2'], 'environment': 'prod'},
                'prod-host5': {'datacenter': 'ONE', 'network': 'DMZ', 'pool': ['hostgroup2'], 'environment': 'prod'},
                'dev-host2': {'datacenter': 'ONE', 'network': 'DEV', 'pool': ['hostgroup2'], 'environment': 'dev'},
                'prod-host4': {'datacenter': 'TWO', 'network': 'DMZ', 'pool': ['hostgroup1'], 'environment': 'prod'},
                'prod-host9': {'datacenter': 'ONE', 'network': 'SEC', 'pool': ['hostgroup4'], 'environment': 'prod'}},
            'application_properties': {
                'application2': {'properties': ['propfile1']},
                'application3': {'properties': ['propfile1', 'propfile2', 'propfile3', 'propfile4']},
                'application1': {'properties': ['propfile1', 'propfile2']}
            },
            "application_hosts": {'application1': {'pool': ['hostgroup1', 'hostgroup2']},
                                  'application3': {'pool': ['hostgroup3', 'hostgroup2']},
                                  'application2': {'pool': ['hostgroup3', 'hostgroup4']}}
        }

        self.data.targets = {
            "hosts": ["pre-host1", "pre-host2", "staging-host1", "staging-host2", "prod-host1", "prod-host2"],
            "applications": ["application1", "application2", "application3"]
        }

    def test_find_dict_diff_with_two_dicts_list_values(self):
        dict1 = {"application_hosts": {'application1': {'pool': ['hostgroup1', 'hostgroup2']},
                                       'application3': {'pool': ['hostgroup3', 'hostgroup2']},
                                       'application2': {'pool': ['hostgroup3', 'hostgroup4']}}}
        dict2 = {"application_hosts": {'application2': {'pool': ['hostgroup3', 'hostgroup2']}}}
        returns = FilterTargets.find_dict_diff(d1=dict1, d2=dict2)
        self.assertTrue(returns)

    def test_find_dict_diff_with_two_dicts_list_values_no_match(self):
        dict1 = {"application_hosts": {'application1': {'pool': ['hostgroup1', 'hostgroup2']},
                                       'application3': {'pool': ['hostgroup3', 'hostgroup2']},
                                       'application2': {'pool': ['hostgroup3', 'hostgroup4']}}}
        dict2 = {"application_hosts": {'application2': {'pool': ['hostgroup5', 'hostgroup2']}}}
        returns = FilterTargets.find_dict_diff(d1=dict1, d2=dict2)
        self.assertFalse(returns)

    def test_find_dict_diff_with_two_dicts_string_values(self):
        dict1 = {"application_hosts": {'application1': {'pool': "test"},
                                       'application3': {'pool': ['hostgroup3', 'hostgroup2']},
                                       'application2': {'pool': "test"}}}
        dict2 = {"application_hosts": {'application1': {'pool': "test"}}}
        returns = FilterTargets.find_dict_diff(d1=dict1, d2=dict2)
        self.assertTrue(returns)

    def test_find_dict_diff_with_two_dicts_string_values_no_match(self):
        dict1 = {"application_hosts": {'application1': {'pool': "test"},
                                       'application3': {'pool': ['hostgroup3', 'hostgroup2']},
                                       'application2': {'pool': "test"}}}
        dict2 = {"application_hosts": {'application1': {'pool': "tester"}}}
        returns = FilterTargets.find_dict_diff(d1=dict1, d2=dict2)
        self.assertFalse(returns)

    def test_FilterHostByHost_exists(self):
        result = False
        conf = dict
        host_name = "pre-host2"
        args = {"data": self.data,
                "conf": conf,
                "host_name": host_name}
        instance = FilterTargets.FilterHostsByHost()
        returns = instance.run(**args)
        if isinstance(returns, object) \
                and isinstance(returns.targets["hosts"], list) \
                and len(returns.targets["hosts"]) == 1 \
                and returns.targets["hosts"][0] == host_name:
            result = True
        self.assertTrue(result)

    def test_FilterHostByHost_not_exists(self):
        result = False
        conf = dict
        host_name = "pre-host5"
        args = {"data": self.data,
                "conf": conf,
                "host_name": host_name}
        instance = FilterTargets.FilterHostsByHost()
        returns = instance.run(**args)
        if isinstance(returns, object) \
                and isinstance(returns.targets["hosts"], list) \
                and len(returns.targets["hosts"]) == 0:
            result = True
        self.assertTrue(result)

    def test_FilterApplicationsByHost_exists(self):
        result = False
        conf = dict
        host_name = "pre-host1"
        args = {"data": self.data,
                "conf": conf,
                "host_name": host_name}
        instance = FilterTargets.FilterApplicationsByHost()
        returns = instance.run(**args)
        if isinstance(returns, object) \
                and isinstance(returns.targets["applications"], list) \
                and len(returns.targets["applications"]) == 1 \
                and returns.targets["applications"][0] == "application1":
            result = True
        self.assertTrue(result)

    def test_FilterApplicationsByHost_not_exists(self):
        result = False
        conf = dict
        host_name = "pre-host123"
        args = {"data": self.data,
                "conf": conf,
                "host_name": host_name}
        instance = FilterTargets.FilterApplicationsByHost()
        returns = instance.run(**args)
        if isinstance(returns, object) \
                and isinstance(returns.targets["applications"], list) \
                and len(returns.targets["applications"]) == 0:
            result = True
        self.assertTrue(result)

    def test_FilterHostsByTag_exists(self):
        result = False
        conf = dict
        tag_name = "pool"
        tag_value = "hostgroup1"
        expected = ['staging-host1', 'prod-host1', 'prod-host2', 'pre-host1']
        args = {"data": self.data,
                "conf": conf,
                "tag_name": tag_name,
                "tag_value": tag_value}
        instance = FilterTargets.FilterHostsByTag()
        returns = instance.run(**args)
        if isinstance(returns, object) \
                and isinstance(returns.targets["hosts"], list) \
                and len(returns.targets["hosts"]) > 0 \
                and set(returns.targets["hosts"]) == set(expected):
            result = True
        self.assertTrue(result)

    def test_FilterHostsByTag_not_exists(self):
        result = False
        conf = dict
        tag_name = "pool"
        tag_value = "hostgroup15"
        args = {"data": self.data,
                "conf": conf,
                "tag_name": tag_name,
                "tag_value": tag_value}
        instance = FilterTargets.FilterHostsByTag()
        returns = instance.run(**args)
        if isinstance(returns, object) \
                and isinstance(returns.targets["hosts"], list) \
                and len(returns.targets["hosts"]) == 0:
            result = True
        self.assertTrue(result)

    def test_FilterHostsByApplication_exists(self):
        result = False
        conf = dict
        application_name = "application1"
        expected = ['pre-host1', 'pre-host2', 'staging-host1', 'staging-host2', 'prod-host1', 'prod-host2']
        args = {"data": self.data,
                "conf": conf,
                "application_name": application_name}
        instance = FilterTargets.FilterHostsByApplication()
        returns = instance.run(**args)
        print(returns.targets["hosts"])
        if isinstance(returns, object) \
                and isinstance(returns.targets["hosts"], list) \
                and len(returns.targets["hosts"]) > 0 \
                and set(returns.targets["hosts"]) == set(expected):
            result = True
        self.assertTrue(result)

    def test_FilterHostsByApplication_not_exists(self):
        result = False
        conf = dict
        application_name = "application12"
        args = {"data": self.data,
                "conf": conf,
                "application_name": application_name}
        instance = FilterTargets.FilterHostsByApplication()
        returns = instance.run(**args)
        print(returns.targets["hosts"])
        if isinstance(returns, object) \
                and isinstance(returns.targets["hosts"], list) \
                and len(returns.targets["hosts"]) == 0:
            result = True
        self.assertTrue(result)

    def test_FilterApplicationsByTag_exists(self):
        result = False
        conf = dict
        tag_name = "pool"
        tag_value = "hostgroup2"
        expected = ['application1', 'application3']
        args = {"data": self.data,
                "conf": conf,
                "tag_name": tag_name,
                "tag_value": tag_value}
        instance = FilterTargets.FilterApplicationsByTag()
        returns = instance.run(**args)
        if isinstance(returns, object) \
                and isinstance(returns.targets["applications"], list) \
                and len(returns.targets["applications"]) > 0 \
                and set(returns.targets["applications"]) == set(expected):
            result = True
        self.assertTrue(result)

    def test_FilterApplicationsByTag_not_exists(self):
        result = False
        conf = dict
        tag_name = "pool"
        tag_value = "hostgroup24"
        args = {"data": self.data,
                "conf": conf,
                "tag_name": tag_name,
                "tag_value": tag_value}
        instance = FilterTargets.FilterApplicationsByTag()
        returns = instance.run(**args)
        if isinstance(returns, object) \
                and isinstance(returns.targets["applications"], list) \
                and len(returns.targets["applications"]) == 0:
            result = True
        self.assertTrue(result)

    def test_FilterApplicationsByApplication_exists(self):
        result = False
        conf = dict
        application_name = "application1"
        args = {"data": self.data,
                "conf": conf,
                "application_name": application_name}
        instance = FilterTargets.FilterApplicationsByApplication()
        returns = instance.run(**args)
        if isinstance(returns, object) \
                and isinstance(returns.targets["applications"], list) \
                and len(returns.targets["applications"]) == 1 \
                and returns.targets["applications"][0] == application_name:
            result = True
        self.assertTrue(result)

    def test_FilterApplicationsByApplication_not_exists(self):
        result = False
        conf = dict
        application_name = "application12"
        args = {"data": self.data,
                "conf": conf,
                "application_name": application_name}
        instance = FilterTargets.FilterApplicationsByApplication()
        returns = instance.run(**args)
        if isinstance(returns, object) \
                and isinstance(returns.targets["applications"], list) \
                and len(returns.targets["applications"]) == 0:
            result = True
        self.assertTrue(result)

    def test_FilterApplicationsByProperty_exists(self):
        result = False
        conf = dict
        expected = ['application1', 'application2', 'application3']
        property_name = "propfile1"
        args = {"data": self.data,
                "conf": conf,
                "property_name": property_name}
        instance = FilterTargets.FilterApplicationsByProperty()
        returns = instance.run(**args)
        print(returns.targets["applications"])
        if isinstance(returns, object) \
                and isinstance(returns.targets["applications"], list) \
                and len(returns.targets["applications"]) > 0 \
                and returns.targets["applications"] == expected:
            result = True
        self.assertTrue(result)

    def test_FilterApplicationsByProperty_not_exists(self):
        result = False
        conf = dict
        expected = ['application1', 'application2', 'application3']
        property_name = "propfile123"
        args = {"data": self.data,
                "conf": conf,
                "property_name": property_name}
        instance = FilterTargets.FilterApplicationsByProperty()
        returns = instance.run(**args)
        print(returns.targets["applications"])
        if isinstance(returns, object) \
                and isinstance(returns.targets["applications"], list) \
                and len(returns.targets["applications"]) == 0:
            result = True
        self.assertTrue(result)



# Todo: Validators
# Todo: Readers
# Todo: PreFilters
# Todo: Resolvers
# # Todo: Hierarchies
# Todo: PostFilters
# # Todo: Placeholders


"""
validators: returns key: boolean (dict)
readers: takes previous data and returns all data (dict)
pre-filters: takes all data, and returns all data (dict)
resolvers: uses hierarchy to resolve data, returns all resolved data (dict)
post-filters: takes all resolved data and returns all filtered data
writers: takes all take and outputs it according to given class(es)

##
filters could be considered one unified type?
## all other than validators, should take and return a dict() containing all data
"""


# -*- coding: utf-8 -*-

from pyconverge.plugins.BasePlugin import BasePlugin
from .LoadTargets import LoadHierarchy, LoadHosts, LoadApplications
from .Filters import FilterApplicationsByHost
from .LoadProperties import LoadProperties
from .OutputProperties import OutputProperties
import logging


class PropertiesPlugin(BasePlugin):
    def __init__(self, **kwargs):
        self.hierarchy_path = kwargs.get("hierarchy")
        self.target_path = kwargs.get("target")
        self.repository_path = kwargs.get("data")
        self.hierarchy = list()
        self.targets = dict()
        self.output = OutputProperties(output_path=kwargs.get("output"))

    def read_hierarchy(self):
        hiera_loader = LoadHierarchy()
        self.hierarchy = hiera_loader.load_contents_of_files(base_directory=self.hierarchy_path)

    def read_targets(self):
        targets = dict()
        # load hosts
        hosts = LoadHosts()
        targets["hosts"] = hosts.load_contents_of_files(base_directory=self.target_path)
        # load applications
        applications = LoadApplications()
        targets["applications"] = applications.load_contents_of_files(base_directory=self.target_path)
        self.targets = targets

    def read_repository(self):
        repository = dict()
        return repository

    def resolve_all_data(self, periodic_write=False):
        resolved_data = dict()

        for host in self.targets["hosts"]:
            resolved_data[host] = self.resolve_target_data(target_name=host, periodic_write=periodic_write)

        return resolved_data

    def resolve_target_data(self, target_name, periodic_write=False):
        resolved_data = dict()
        filters = FilterApplicationsByHost()
        host_tags = self.targets["hosts"][target_name]
        filtered_applications = filters.get_applications_matching_host(applications=self.targets["applications"],
                                                                       host_tags=host_tags)
        lines = 0
        for application in filtered_applications:
            application_tags = self.targets["applications"][application]
            property_loader = LoadProperties(hierarchy=self.hierarchy,
                                             host_tags=host_tags,
                                             repository_path=self.repository_path)
            result = property_loader.load_contents_of_property_list(application_name=application,
                                                                    application_tags=application_tags)
            if result:
                lines += len(result)
                resolved_data[application] = result

        logging.info("Generating data for target: %-15s (files: %-3i, lines: %3i)" % (target_name,
                                                                                len(resolved_data),
                                                                                lines))
        if periodic_write:
            self.write_data(resolved_data=resolved_data)

        return resolved_data

    def write_data(self, resolved_data):
        return self.output.generate_files_by_target(configuration=resolved_data)



# -*- coding: utf-8 -*-


def find_dict_diff(d1, d2, path=""):
    result = False
    for k in d1.keys():
        if k not in d2:
            continue
        elif isinstance(d1[k], dict):
            if path == "":
                path = k
            else:
                path = path + "->" + k
            result = find_dict_diff(d1[k], d2[k], path)
        elif d1[k] == d2[k] or \
                (isinstance(d1[k], list)
                 and isinstance(d2[k], list)
                 and len(set(d1[k]).intersection(set(d2[k]))) > 0):
            return True
    return result


class FilterHostsByHost:
    @staticmethod
    def run(data, conf, **kwargs):
        data_filter = kwargs.get("host_name")
        filtered_targets = list()
        if data_filter in data.targets["hosts"]:
            filtered_targets.append(data_filter)
        data.targets["hosts"] = filtered_targets
        return data


class FilterHostsByTag:
    @staticmethod
    def run(data, **kwargs):
        tag_name = kwargs.get("tag_name")
        tag_value = kwargs.get("tag_value")
        filtered_targets = list()
        for host_name in data.targets["hosts"]:
            host_tags = data.data["hosts"][host_name]
            if tag_name in host_tags and (
                        (isinstance(host_tags[tag_name], str) and host_tags[tag_name] == tag_value) or
                        (isinstance(host_tags[tag_name], list) and any(
                                host_value == tag_value for host_value in host_tags[tag_name]))):
                filtered_targets.append(host_name)
        data.targets["hosts"] = filtered_targets
        return data


class FilterHostsByApplication:
    @staticmethod
    def run(data, **kwargs):
        application_name = kwargs.get("application_name")
        filtered_data = list()
        if application_name in data.targets["applications"]:
            application_tags = data.data["application_hosts"][application_name]
            for host_name in data.targets["hosts"]:
                host_tags = data.data["hosts"][host_name]
                host_app_tag_match = find_dict_diff(host_tags, application_tags)
                if host_app_tag_match:
                    filtered_data.append(host_name)
        data.targets["hosts"] = filtered_data
        return data


class FilterApplicationsByTag:
    @staticmethod
    def run(data, **kwargs):
        tag_name = kwargs.get("tag_name")
        tag_value = kwargs.get("tag_value")
        filtered_targets = list()
        for application_name in data.targets["applications"]:
            application_tags = data.data["application_hosts"][application_name]
            if tag_name in application_tags and (
                        (isinstance(application_tags[tag_name], str)
                         and application_tags[tag_name] == tag_value) or
                        (isinstance(application_tags[tag_name], list)
                         and any(app_value == tag_value for app_value in application_tags[tag_name]))):
                filtered_targets.append(application_name)
        data.targets["applications"] = filtered_targets
        return data


class FilterApplicationsByHost:
    @staticmethod
    def run(data, **kwargs):
        host_name = kwargs.get("host_name")
        filtered_data = list()
        if host_name in data.data["hosts"]:
            host_tags = data.data["hosts"][host_name]
            for application_name in data.targets["applications"]:
                application_tags = data.data["application_hosts"][application_name]
                app_host_tag_match = find_dict_diff(application_tags, host_tags)
                if app_host_tag_match:
                    filtered_data.append(application_name)
        data.targets["applications"] = filtered_data
        return data


class FilterApplicationsByApplication:
    @staticmethod
    def run(data, **kwargs):
        data_filter = kwargs.get("application_name")
        filtered_targets = list()
        if data_filter in data.targets["applications"]:
            filtered_targets.append(data_filter)
        data.targets["applications"] = filtered_targets
        return data


class FilterApplicationsByProperty:
    @staticmethod
    def run(data, **kwargs):
        property_name = kwargs.get("property_name")
        filtered_targets = list()
        for application_name in data.targets["applications"]:
            application_props = data.data["application_properties"][application_name]
            if "properties" in application_props and \
                    isinstance(application_props["properties"], list) and \
                            property_name in application_props["properties"]:
                filtered_targets.append(application_name)
        data.targets["applications"] = filtered_targets
        return data



# -*- coding: utf-8 -*-


class FilterApplicationsByHost:
    @staticmethod
    def get_applications_matching_host(applications, host_tags):
        filtered_applications = list()
        for application_name, application_values in applications.items():
            for tag_type, tag_values in host_tags.items():
                if tag_type in application_values:
                    for tag_value in tag_values:
                        if any(x == tag_value for x in application_values[tag_type]):
                            filtered_applications.append(application_name)
        return filtered_applications



import glob
import os
import yaml
from abc import ABC, abstractmethod
import re
import configparser
from itertools import chain


class LoadDataFromDisk(ABC):
    @staticmethod
    def get_list_of_files(glob_pattern, recursive=False):
        return glob.glob(glob_pattern, recursive=recursive)

    @abstractmethod
    def merge_contents_of_files(self, file_list):
        pass

    @abstractmethod
    def load_contents_of_files(self):
        pass


class LoadHosts(LoadDataFromDisk):
    def merge_contents_of_files(self, file_list):
        contents = dict()
        all_yaml_individual = map(lambda x: yaml.load_all(open(x)), file_list)
        for configs in all_yaml_individual:
            for config in configs:
                contents = {**contents, **config}
        return contents

    def load_contents_of_files(self):
        glob_pattern = os.path.join("targets", "hosts", "*.yaml")
        file_list = self.get_list_of_files(glob_pattern=glob_pattern, recursive=False)
        content = self.merge_contents_of_files(file_list=file_list)
        return content


class LoadApplications(LoadDataFromDisk):
    def merge_contents_of_files(self, file_list):
        contents = dict()
        for file_path in file_list:
            app_name = file_path.rsplit("/", 1)[1][:-5]
            for config in yaml.load_all(open(file_path)):
                contents[app_name] = config
        return contents

    def load_contents_of_files(self):
        glob_pattern = os.path.join("targets", "applications", "*.yaml")
        file_list = self.get_list_of_files(glob_pattern=glob_pattern, recursive=False)
        content = self.merge_contents_of_files(file_list=file_list)
        return content


class LoadHierarchy(LoadDataFromDisk):
    def merge_contents_of_files(self, file_list):
        contents = list()
        all_yaml_individual = map(lambda x: yaml.load_all(open(x)), file_list)
        for configs in all_yaml_individual:
            for config in configs:
                contents.extend(config)
        return contents

    def load_contents_of_files(self):
        glob_pattern = os.path.join("hierarchy", "hierarchy.yaml")
        file_list = self.get_list_of_files(glob_pattern=glob_pattern, recursive=False)
        content = self.merge_contents_of_files(file_list=file_list)
        return content


class FilterApplicationsByHost:
    @staticmethod
    def get_applications_matching_host(applications, host_tags, host_name):
        # print(applications)
        filtered_applications = list()
        for application_name, application_values in applications.items():
            # print(host_name, host_tags)
            for tag_type, tag_values in host_tags.items():
                if tag_type in application_values:
                    # print("tag: %s, values: %s" % (tag_type, tag_values))
                    # print(tag_type in application_values, tag_type, application_values)
                    for tag_value in tag_values:
                        # print(tag_value in tag_values, tag_value, tag_values, application_values[tag_type])
                        if any(x == tag_value for x in application_values[tag_type]):
                            print("GOTYA! %s" % application_name)
                            filtered_applications.append(application_name)
        return filtered_applications


class LoadProperties(LoadDataFromDisk):
    def __init__(self, hierarchy, host_tags):
        self.host_tags = host_tags
        self.hierarchy = hierarchy
        self.regex_hierarchy = self.get_regex_hierarchy(hierarchy=hierarchy)

    def merge_contents_of_files(self, file_list):
        parser = configparser.ConfigParser(allow_no_value=True)
        for file_path in file_list:
            with open(file_path) as lines:
                lines = chain(("[default]",), lines)  # This line does the trick.
                parser.read_file(lines)

        contents = parser.items("default")
        return contents

    def load_contents_of_files(self):
        pass

    @staticmethod
    def get_regex_hierarchy(hierarchy):
        regex_hierarchy = []
        for hiera in hierarchy:
            regex_hiera = dict()
            regex_hiera['hiera'] = hiera
            regex_hiera['regex'] = "data/" + re.sub("\${.+?}", "([^/]+)", hiera)
            regex_hiera['tags'] = re.findall("\$\{([^/]+)\}", hiera)
            regex_hierarchy.append(regex_hiera)
        return regex_hierarchy

    def filter_hierarchy(self, file_list):
        filtered_hierarchy = list()

        for file_path in file_list:
            for hiera in self.regex_hierarchy:
                if re.match(hiera['regex'], os.path.dirname(file_path)):
                    filtered_hierarchy.append(hiera)
        return filtered_hierarchy

    def filter_file_list_by_host_tags(self, file_list, filtered_hierarchy, application_name):
        content = list()
        # print("START")
        for file_path in file_list:
            from pprint import pprint

            for hiera in filtered_hierarchy:
                regex = re.findall(hiera["regex"], os.path.dirname(file_path))

                if regex:
                    result = True
                    for i, tag in enumerate(regex):
                        if not hiera['tags'] and tag.rsplit("/", 1)[1] == "default":
                            break
                        elif hiera['tags'][i] == "app":
                            if application_name != tag:
                                result = False
                                break
                        elif hiera['tags'][i] in self.host_tags \
                                and hiera['tags'][i] in self.host_tags:
                            break
                        else:
                            result = False
                            break
                    if result:
                        content.append(file_path)
        return content

    def get_file_list_ordered_by_hierarchy(self, file_list, application_name, application_tags):
        content = list()
        hierarchy = self.filter_hierarchy(file_list)
        filtered_file_list = self.filter_file_list_by_host_tags(file_list=file_list,
                                                                filtered_hierarchy=hierarchy,
                                                                application_name=application_name)
        if len(filtered_file_list) > 0:
            for hiera in hierarchy:
                for file_path in filtered_file_list:
                    if re.match(hiera['regex'], file_path):
                        content.append(file_path)
                        filtered_file_list.remove(file_path)
        return content

    def load_contents_of_property_by_pattern(self, file_name, application_name, application_tags):
        content = list()
        glob_pattern = os.path.join("data", "**", file_name + ".properties")
        file_list = self.get_list_of_files(glob_pattern=glob_pattern, recursive=True)
        file_list_order_by_hierarchy = self.get_file_list_ordered_by_hierarchy(file_list=file_list,
                                                                               application_name=application_name,
                                                                               application_tags=application_tags)
        if file_list_order_by_hierarchy:
            content = self.merge_contents_of_files(file_list=file_list_order_by_hierarchy)
        return content

    def load_contents_of_property_list(self, application_name, application_tags):
        content = dict()
        for file_path in application_tags['properties']:
            result = self.load_contents_of_property_by_pattern(file_name=file_path,
                                                               application_name=application_name,
                                                               application_tags=application_tags)

            # if result:
            content[file_path] = result
        return content


class OutputProperties:

    def generate_file_by_host(self, configuration):
        result = False
        output_folder = os.path.join("output")
        for host_name, applications in configuration.items():
            for application_name, properties in applications.items():
                output_folder = os.path.join("output", host_name, application_name)
                if not os.path.exists(output_folder):
                    os.makedirs(output_folder)
                for property_name, keys in properties.items():
                    file_name = "%s.properties" % property_name

                    for key in keys:
                        with open(os.path.join(output_folder, file_name), 'w+') as out:
                            out.write("%s = %s" % (key[0], key[1]))

        return result

def main():
    hosts = LoadHosts().load_contents_of_files()
    applications = LoadApplications().load_contents_of_files()
    hierarchy = LoadHierarchy().load_contents_of_files()
    fa = FilterApplicationsByHost()

    all_properties = dict()
    processed = []
    for host_name, host_tags in hosts.items():
        print("\n\nHOST: ", host_name)
        # print("host tags: ", host_tags)
        filtered_applications = fa.get_applications_matching_host(applications=applications, host_tags=host_tags, host_name=host_name)
        print("FILTER", filtered_applications)
        host_properties = dict()
        processed.append(host_name)

        for application_name in filtered_applications:
            application_tags = applications[application_name]
            property_loader = LoadProperties(hierarchy=hierarchy, host_tags=host_tags)
            result = property_loader.load_contents_of_property_list(application_name=application_name,
                                                                    application_tags=application_tags)
            # if result:
            host_properties[application_name] = result

        # if host_properties:
        all_properties[host_name] = host_properties

    output = OutputProperties()
    output.generate_file_by_host(configuration=all_properties)
    from pprint import pprint
    # pprint(all_properties)


if __name__ == '__main__':
    main()



from .LoadDataFromDisk import LoadDataFromDisk
import yaml
import os


class Hierarchy(LoadDataFromDisk):

    def __init__(self, settings):
        self.conf = settings["yaml"]

    def merge_contents_of_files(self, file_list):
        contents = list()
        all_yaml_individual = map(lambda x: yaml.load_all(open(x)), file_list)
        for configs in all_yaml_individual:
            for config in configs:
                contents.extend(config)
        return contents

    def read_data(self):
        glob_pattern = os.path.join(self.conf["base_dir"], self.conf["hierarchy_path"])
        file_list = self.get_list_of_files(glob_pattern=glob_pattern, recursive=False)
        content = self.merge_contents_of_files(file_list=file_list)
        return content





# -*- coding: utf-8 -*-


"""setup.py: setuptools control."""

import re
from setuptools import setup
import os


def read_md(f):
    try:
        from pypandoc import convert
        return convert(f, 'rst')
    except ImportError:
        print("warning: pypandoc module not found, could not convert Markdown to RST")
        return open(f, 'r').read()


def package_files(directory):
    paths = ()
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths


version = re.search(
    '^__version__\s*=\s*"(.*)"',
    open('pyconverge/__init__.py').read(),
    re.M
).group(1)


setup(
    name="pyconverge",
    packages=["pyconverge", "tests", "pyconverge/resources", "pyconverge/schemas", *[x[0] for x in os.walk("pyconverge/plugins")]],
    entry_points={
        "console_scripts": ['converge = pyconverge.converge:main']
    },
    license='Apache License 2.0',
    version=version,
    install_requires=['pyyaml', 'pytest', 'pytest-cov', 'pykwalify'],
    description="Resolve configurations from abstract hierarchies and templates",
    long_description=read_md("README.md"),
    author="Andrew Boswell",
    author_email="drewboswell@gmail.com",
    url="https://github.com/drewboswell/converge",
    platforms='any',
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Development Status :: 1 - Planning',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Financial and Insurance Industry',
        'Operating System :: OS Independent',
        'Topic :: Office/Business'
    ],
    keywords='configuration management development operations system sysadmin config converge',
    include_package_data=True,
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
)



# -*- coding: utf-8 -*-

from .LoadDataFromDisk import LoadDataFromDisk
import configparser
from itertools import chain
import logging
import re
import os


class LoadProperties(LoadDataFromDisk):
    def __init__(self, hierarchy, host_tags, repository_path):
        self.host_tags = host_tags
        self.hierarchy = hierarchy
        self.repository_path = repository_path
        self.regex_hierarchy = self.get_regex_hierarchy(hierarchy=hierarchy)

    def merge_contents_of_files(self, file_list):
        parser = configparser.ConfigParser(allow_no_value=True)
        for file_path in file_list:
            with open(file_path) as lines:
                lines = chain(("[default]",), lines)  # This line does the trick...
                parser.read_file(lines)

        contents = parser.items("default")
        return contents

    def load_contents_of_files(self):
        pass

    def get_regex_hierarchy(self, hierarchy):
        regex_hierarchy = []
        for hiera in hierarchy:
            regex_hiera = dict()
            regex_hiera['hiera'] = hiera
            regex_hiera['regex'] = os.path.join(self.repository_path, re.sub("\${.+?}", "([^/]+)", hiera))
            regex_hiera['tags'] = re.findall("\$\{([^/]+)\}", hiera)
            regex_hierarchy.append(regex_hiera)
        return regex_hierarchy

    def filter_hierarchy(self, file_list):
        filtered_hierarchy = list()
        for file_path in file_list:
            for hiera in self.regex_hierarchy:
                if re.match(hiera['regex'], os.path.dirname(file_path)):
                    filtered_hierarchy.append(hiera)
        return filtered_hierarchy

    def filter_file_list_by_host_tags(self, file_list, filtered_hierarchy, application_name):
        content = list()
        for file_path in file_list:
            for hiera in filtered_hierarchy:
                regex = re.findall(hiera["regex"], os.path.dirname(file_path))
                if regex:
                    result = True
                    for i, tag in enumerate(regex):
                        if (not hiera['tags'] and tag.rsplit("/", 1)[1] == "default") \
                                or hiera['tags'][i] in self.host_tags:
                            break
                        elif hiera['tags'][i] == "app":
                            if application_name != tag:
                                result = False
                                break
                        else:
                            result = False
                            break
                    if result:
                        content.append(file_path)
        return content

    def get_file_list_ordered_by_hierarchy(self, file_list, application_name, application_tags):
        content = list()
        hierarchy = self.filter_hierarchy(file_list)
        filtered_file_list = self.filter_file_list_by_host_tags(file_list=file_list,
                                                                filtered_hierarchy=hierarchy,
                                                                application_name=application_name)
        if len(filtered_file_list) > 0:
            for hiera in hierarchy:
                for file_path in filtered_file_list:
                    if re.match(hiera['regex'], file_path):
                        content.append(file_path)
                        filtered_file_list.remove(file_path)
        return content

    def load_contents_of_property_by_pattern(self, file_name, application_name, application_tags):
        content = list()
        logging.info("Loading Contents for %s" % application_name)
        glob_pattern = os.path.join(self.repository_path, "**", file_name + ".properties")
        file_list = self.get_list_of_files(glob_pattern=glob_pattern, recursive=True)
        file_list_order_by_hierarchy = self.get_file_list_ordered_by_hierarchy(file_list=file_list,
                                                                               application_name=application_name,
                                                                               application_tags=application_tags)
        if file_list_order_by_hierarchy:
            content = self.merge_contents_of_files(file_list=file_list_order_by_hierarchy)
        return content

    def load_contents_of_property_list(self, application_name, application_tags):
        content = dict()
        import time
        start = time.time()
        for file_path in application_tags['properties']:
            result = self.load_contents_of_property_by_pattern(file_name=file_path,
                                                               application_name=application_name,
                                                               application_tags=application_tags)
            if result:
                content[file_path] = result
        logging.info("TIMING load_contents_of_property_list %s %f" % (application_name, time.time() - start))
        return content



# -*- coding: utf-8 -*-

from abc import ABC, abstractmethod


class BasePlugin(ABC):

    def read_data(self):
        result = dict()
        result["hierarchy"] = self.read_hierarchy
        result["targets"] = self.read_targets
        result["repository"] = self.read_repository
        return result

    @abstractmethod
    def read_hierarchy(self):
        pass

    @abstractmethod
    def read_targets(self):
        pass

    @abstractmethod
    def read_repository(self):
        pass

    @abstractmethod
    def resolve_all_data(self, periodic_write=False):
        pass

    @abstractmethod
    def resolve_target_data(self, target, periodic_write=False):
        pass

    @abstractmethod
    def write_data(self, resolved_data):
        pass
